from .errors import *
from typing import Callable, Sequence
from inspect import signature
import asyncio


class CustomConverter:
    """
    Our custom converter class which others need to inherit from.
    ```
    MAKE SURE TO DEFINE __new__ IN THIS!
    ```
    """

    def __init__(self, arg: str):
        self.arg = arg

    def __new__(cls, arg: str):  # Make sure to define this in your own converter
        raise NotImplementedError(
            "The subclasses for converters need to define __new__ and return the values"
        )


class Command:
    """
    Our base command class from which commands are formed.
    """

    def __init__(
        self,
        name: str,
        aliases: Sequence[str],
        func: Callable,
        *,
        usage: str = None,
        description: str
    ):
        self.args: list[str] = [
            i.strip() for i in str(signature(func))[1:-1].split(", ")
        ]
        self.func = func
        self.name = name or self.func.__name__
        self.aliases = aliases
        self.signature = "[" + "] [".join(self.args) + "]"
        self.usage = usage or self.signature
        self.description = description
        for idx, arg in enumerate(self.args):
            if ":" in arg:
                self.args[idx] = (arg.split)(":")[0].strip()
            elif "=" in arg:
                self.args[idx] = arg.split("=")[0].strip()

        for idx, arg in enumerate(self.args):
            pass

    def invoke(self, command_str: str):
        command_li: list[str] = command_str.split(" ")
        for idx, arg in enumerate(command_li):
            if not arg:
                continue

            if any("*" in arg for arg in self.args):
                _type = str
            else:
                _type = self.func.__annotations__.get(self.args[idx]) or str
            try:
                command_li[idx] = _type(arg)
            except ValueError as ba:
                raise BadArgument(*ba.args)
            if "" in command_li:
                command_li.remove("")
        while "" in command_li:
            command_li.remove("")
        if asyncio.iscoroutinefunction(self.func):
            asyncio.run(self.func(*command_li))
        else:
            self.func(*command_li)



class CommandError(Exception):
    """
    Our command Invoking error that each error needs to derive from.
    """

    pass


class CommandInvokeError(CommandError):
    """
    An exception raised on errors in invoking of a command.
    """

    pass


class BadArgument(CommandInvokeError):
    """
    An exception raised on a bad provided argument.
    """

    pass


class CommandCreateError(CommandError):
    """
    An exception which is triggered on an error in creation of a command.
    """

    pass



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="python-commandmaker",
    version="0.0.4",
    author="Cube Riser",
    author_email="cuberiser@gmail.com",
    long_description_content_type="text/markdown",
    long_description=str(long_description),
    description="Python module for discord bot like commands from the console",
    python_requires=">=3.10.0",
    packages=setuptools.find_packages(),
    install_requires=[""],
)



"""
Check https://pypi.org/project/python-commandmaker, https://github.com/cuberiser/python-commandmaker for more info
"""
from typing import Callable, Sequence
from .errors import *
from .core import Command
import sys


__version__ = "0.0.4"
__author__ = "Cube Riser"


class CommandMaker:
    """
    This is our base command maker class from which we make commands.
    You can either inherit from this or make an instance to create commands.
    """

    def __init__(
        self, name: str, prefix: str | Callable, *, case_insensitive: bool, **kwargs
    ):
        self._cs = case_insensitive
        self._prefix = prefix
        self.__name__ = name
        self._commands: dict[str, Command] = {}
        self._aliases: dict[str, Command] = {}
        for kwarg in kwargs:
            self.__setattr__(kwarg, kwargs[kwarg])

    def command(
        self,
        *,
        name: str = "",
        aliases: Sequence[str] = [],
        description: str = "No description",
        override=False,
    ):
        if not override:
            if self.get_command(name):
                raise CommandCreateError(f"Already an existing command or alias {name}")
            for alias in aliases:
                if self.get_command(alias):
                    raise CommandCreateError(
                        f"Already an existing command or alias {alias}"
                    )
        def _wrapper(func: Callable):
            cmd = Command(name, aliases, func, description=description)
            self._commands[name or func.__name__] = cmd
            for a in aliases:
                self._aliases[str(a)] = cmd
            if override:
                print(f"WARNING: Overriding a command might have unintentional effects in some commands\ncommand:{name or func.__name__}", file=sys.stderr)

        return _wrapper

    def get_command(self, command_name: str):
        return self._commands.get(command_name) or self._aliases.get(command_name)

    @property
    def prefix(self) -> str:
        if callable(self._prefix):
            return self._prefix(self)
        return str(self._prefix)

    @property
    def commands(self):
        for command in self._commands.values():
            yield command

    def run(self):
        print(
            f"Welcome to {self.__name__}'s commands, my prefix right now is {self.prefix}"
        )
        while True:
            command = input(">>> ") if not self._cs else input(">>> ").lower()
            if command.startswith(self.prefix):
                command = command[len(self.prefix) :]
                cmd = self.get_command(command.split(" ")[0])
                if cmd:
                    try:
                        cmd.invoke(" ".join(command.split(" ")[1:]))
                    except ValueError as e:
                        print("A value error occured", e)
                    except BadArgument as ba:
                        print("Bad argument:", ba)
                    except TypeError as te:
                        print("A type error occured:", te)
                else:
                    print("command not found")
            elif command == "help":
                for command in self._commands.values():
                    print(command.name, command.signature, "\n", command.description)
            elif command == "exit":
                break
            elif command == "prefix":
                print("My current prefix is", self.prefix)
